pub struct Solution;

// Copy from here for submitting this solution to leet code
use std::collections::VecDeque;

impl Solution {
    pub fn can_jump(nums: Vec<i32>) -> bool {
        Solution::idiom_can_jump(&nums)
    }

    /// Returns true if the last element can be reached
    /// by starting with the 0th element in accordance to the rules of the jump game.
    /// Link to the description of this code challenge.
    /// https://leetcode.com/problems/jump-game/description/
    /// Source code of this solution is hosted on:
    /// https://git.kb-one.de/boolpurist/jump_game_challenge_solution/src/branch/main/src/lib.rs
    pub fn idiom_can_jump(nums: &[i32]) -> bool {
        if nums.is_empty() {
            return false;
        }
        // parameter can not be empty at this point
        let target_val = (nums.len() - 1) as i32;
        // Using a dequeue ensures that every index is visited
        // Reason: This algorithm finds every index in an ascending order.
        // Dequeue acts here as a queue preserving the ascending order in which the indices are
        // found
        let mut next_nodes = VecDeque::from([0]);
        // This index and all previous ones are visited already
        let mut greatest_visited_index = 0;
        while let Some(next_index) = next_nodes.pop_front() {
            // Indexing with a next_index greater than the length
            // of parameter nums will not happen
            // Reason: the values for next_index greater than len of nums
            // are not used for indexing because the check for reaching the last index
            // is done before using next_index for indexing
            let jump_height = nums[next_index as usize];
            let max_jump_height = next_index + jump_height;
            let found_end_of_seq = target_val <= max_jump_height;
            if found_end_of_seq {
                return true;
            }
            // Going left to right and using a dequeue for FIFO access,
            // all indices are visited
            let not_visited = greatest_visited_index + 1;
            // It can start at greatest_visited_indices + 1 because the every index was inpected
            // in a previous iterations.
            // If the max_jump_height is lower than the start_point
            // then there is not new index to be discovered
            let last_unvisited = max_jump_height + 1;
            // This loops grantees that every reachable index is visited and only once
            // in the worst case scenario
            // This leads to a linear time complexity for comparison done by Self::check_with_target_value
            for to_inspect_later in not_visited..last_unvisited {
                next_nodes.push_back(to_inspect_later);
                greatest_visited_index = to_inspect_later;
            }
        }
        false
    }
}
// Stop copying from here for submitting this solution to leet code

#[cfg(test)]
mod testing {
    use crate::Solution;

    #[test]
    fn empty_input() {
        let given = vec![];
        let expected = false;
        assert_case(given, expected);
    }

    #[test]
    fn jumping_no_matter_because_one_element() {
        let given = vec![0];
        let expected = true;
        assert_case(given, expected);
    }

    #[test]
    // Example 1:
    // Input: nums = [2,3,1,1,4]
    // Output: true
    // Explanation: Jump 1 step from index 0 to 1, then 3 steps to the last index.
    fn jump_possible() {
        //        Example 1:
        //
        let given = vec![2, 3, 1, 1, 4];
        let expected = true;
        assert_case(given, expected);
    }

    // Example 2:
    //
    // Input: nums = [3,2,1,0,4]
    // Output: false
    #[test]
    fn no_jump_possible() {
        let given = vec![3, 2, 1, 0, 4];
        let expected = false;
        assert_case(given, expected);
    }

    #[test]
    // Taken from a failed submission
    fn jump_two_elements() {
        let given = vec![2, 0];
        let expected = true;
        assert_case(given, expected);
    }

    #[test]
    // Taken from a failed submission
    fn can_jump_complex_array() {
        let given = vec![3, 0, 8, 2, 0, 0, 1];
        let expected = true;
        assert_case(given, expected);
    }
    #[test]
    // Taken from a failed submission
    fn very_complex_cant_jump_complex_array() {
        let given = vec![
            3, 6, 2, 7, 1, 2, 8, 5, 7, 8, 3, 6, 4, 4, 3, 7, 5, 8, 4, 4, 5, 5, 4, 7, 8, 7, 8, 1, 2,
            4, 8, 1, 6, 3, 6, 5, 3, 4, 6, 1, 1, 4, 7, 6, 1, 7, 7, 4, 8, 1, 1, 3, 3, 1, 7, 3, 4, 6,
            2, 5, 8, 4, 1, 8, 4, 6, 4, 7, 5, 5, 6, 2, 2, 5, 2, 7, 5, 1, 7, 5, 8, 1, 1, 6, 4, 6, 8,
            3, 7, 1, 1, 8, 2, 1, 3, 2, 5, 5, 5, 4, 7, 4, 5, 5, 7, 1, 7, 6, 6, 6, 6, 5, 1, 4, 5, 5,
            2, 3, 7, 5, 8, 4, 2, 8, 3, 5, 7, 2, 1, 1, 3, 2, 8, 2, 2, 6, 2, 5, 7, 2, 3, 8, 1, 7, 7,
            3, 2, 4, 6, 6, 5, 1, 5, 7, 2, 1, 6, 4, 4, 5, 3, 4, 8, 2, 7, 4, 2, 1, 2, 1, 4, 8, 2, 3,
            7, 2, 1, 6, 5, 8, 8, 6, 6, 4, 5, 4, 4, 5, 1, 6, 5, 2, 3, 5, 6, 4, 6, 7, 4, 8, 4, 6, 4,
            7, 5, 8, 6, 4, 3, 7, 4, 2, 3, 4, 8, 2, 1, 2, 5, 1, 8, 4, 6, 7, 4, 1, 1, 8, 8, 7, 7, 7,
            3, 1, 2, 7, 2, 4, 5, 8, 5, 8, 2, 8, 3, 3, 6, 5, 5, 6, 3, 2, 5, 8, 5, 4, 6, 2, 5, 8, 7,
            8, 7, 8, 6, 5, 6, 1, 3, 3, 4, 4, 6, 4, 5, 5, 5, 8, 5, 4, 4, 1, 5, 2, 8, 8, 7, 3, 7, 3,
            7, 2, 4, 5, 2, 6, 2, 6, 5, 8, 4, 3, 4, 7, 4, 8, 8, 4, 3, 4, 7, 5, 5, 1, 3, 7, 5, 1, 8,
            4, 4, 2, 7, 5, 3, 1, 5, 8, 7, 3, 3, 6, 1, 1, 7, 1, 3, 8, 7, 1, 2, 4, 3, 5, 2, 2, 5, 1,
            6, 8, 2, 4, 8, 8, 4, 4, 3, 1, 2, 6, 2, 1, 1, 8, 1, 8, 8, 1, 5, 4, 1, 8, 1, 1, 2, 7, 8,
            5, 6, 6, 5, 8, 5, 8, 1, 5, 5, 6, 5, 5, 3, 2, 3, 8, 4, 7, 5, 7, 4, 6, 5, 2, 7, 7, 8, 1,
            7, 2, 4, 6, 5, 1, 3, 1, 8, 1, 5, 2, 2, 6, 6, 6, 1, 4, 8, 4, 1, 3, 7, 1, 1, 1, 7, 4, 5,
            5, 6, 4, 5, 3, 5, 3, 5, 2, 7, 4, 3, 1, 4, 1, 7, 7, 6, 4, 3, 4, 2, 4, 2, 7, 3, 4, 1, 8,
            8, 6, 7, 2, 3, 2, 5, 6, 5, 6, 5, 1, 2, 1, 6, 2, 1, 4, 2, 2, 6, 6, 7, 1, 2, 6, 2, 4, 2,
            4, 1, 6, 7, 5, 5, 4, 5, 4, 4, 2, 4, 8, 6, 8, 1, 6, 3, 7, 7, 8, 4, 2, 8, 1, 2, 1, 7, 8,
            4, 7, 2, 8, 6, 8, 3, 4, 3, 8, 8, 7, 3, 2, 4, 4, 1, 4, 6, 4, 6, 6, 7, 7, 4, 5, 8, 3, 8,
            3, 6, 8, 8, 1, 1, 7, 1, 2, 2, 7, 8, 1, 6, 7, 6, 3, 8, 6, 5, 5, 2, 2, 8, 5, 1, 6, 5, 3,
            6, 4, 2, 8, 5, 3, 1, 3, 4, 4, 1, 5, 2, 4, 8, 3, 7, 4, 1, 4, 7, 8, 4, 1, 3, 8, 3, 7, 4,
            1, 1, 4, 3, 3, 2, 6, 6, 2, 8, 8, 6, 5, 4, 4, 6, 1, 8, 5, 5, 8, 3, 8, 1, 5, 6, 4, 1, 1,
            6, 3, 8, 8, 7, 1, 6, 5, 5, 7, 4, 8, 7, 2, 6, 6, 8, 4, 4, 7, 2, 1, 7, 3, 5, 8, 8, 6, 2,
            2, 5, 1, 3, 7, 7, 3, 4, 8, 5, 3, 7, 3, 3, 4, 2, 1, 4, 2, 6, 1, 1, 4, 7, 6, 8, 2, 3, 5,
            3, 2, 2, 7, 1, 2, 1, 4, 2, 4, 8, 5, 2, 3, 7, 2, 4, 3, 6, 3, 8, 5, 7, 3, 7, 5, 3, 4, 3,
            8, 7, 4, 4, 7, 3, 2, 1, 1, 1, 1, 3, 1, 6, 3, 6, 4, 3, 2, 1, 1, 1, 6, 5, 2, 2, 5, 5, 6,
            6, 1, 8, 7, 8, 7, 6, 2, 2, 3, 5, 6, 1, 0, 0, 0, 0, 0, 0, 0, 0,
        ];
        let expected = false;
        assert_case(given, expected);
    }

    fn assert_case(given: Vec<i32>, expected: bool) {
        let actual = Solution::idiom_can_jump(&given);
        assert_eq!(expected, actual, "Given input: {:#?}", &given);
    }
}
